/*****************************************************************************
* yeehaa -- Control application for v2 or v3 of the Dream Cheeky Rocket Baby
* Copyright (C) 2009 Matt Davis <mattdavis9@gmail.com>
* 
* yeehaa.c
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
*
* Notes
* -----
* This is a userland application to operate the Dream Cheeky USB Missile
* Launcher Version 2.  This application should also operate on their Version 3
* model.  Version 2 of the launcher is also known as "RocketBaby."  Dream
* Cheeky is also Dream Link Ltd.
*
*
* Thanks
* ------
* Some of the USB device codes, shooting algorithm information, status
* query information, and the USB "set_configuration" hint, were obtained from
* Abe Masahiro's Perl library:
* <http://search.cpan.org/~pen/Device-USB-MissileLauncher-RocketBaby-1.01/lib/Device/USB/MissileLauncher/RocketBaby.pm>
*
* Thanks also to Alvin Wong and Dream Link Ltd.'s USB Missile Launcher
* Developer Manual v1.0.
*****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <usb.h>
#include <curses.h>
#include <sys/stat.h>
#include <sys/types.h>


#define YEEHAA_VERSION "0.1"


#define TAG     "[yeehaa]"
#define VENDOR  0x0a81
#define PRODUCT 0x0701


/* Commands (v2 and v3) */
#define IGNORE         -1
#define POSITION_UNSET 0x00
#define POSITION_DOWN  0x01
#define POSITION_UP    0x02
#define POSITION_LEFT  0x04
#define POSITION_RIGHT 0x08
#define SHOOT          0x10
#define STOP           0x20
#define STATUS         0x40
static char _shoot  =  SHOOT;
static char _stop   =  STOP;
static char _status =  STATUS;


/* Status flags */
#define LIMIT_MASK  0x0F
#define DOWN_LIMIT  0x01
#define UP_LIMIT    0x02
#define LEFT_LIMIT  0x04
#define RIGHT_LIMIT 0x08
#define HAS_SHOT    0x10


#define YEEHAA_ERR(_f) { if (!(_f)) { perror(#_f); abort();} }


#define YEEHAA_ERR_MSG(...)                      \
{                                                \
    fprintf(stderr, TAG "[error] " __VA_ARGS__); \
    fputc('\n', stderr);                         \
}


#define DO_CONTROL(_data)                                         \
{                                                                 \
    usb_control_msg(hnd,                                          \
                    0x21,                      /* Request Type */ \
                    USB_REQ_SET_CONFIGURATION, /* Request      */ \
                    0x02,                      /* Value        */ \
                    0x00,                      /* Index        */ \
                    (&_data),                  /* Data         */ \
                    1,                         /* Data Size    */ \
                    1000);                     /* Timeout      */ \
}


/* Callibration structure */
static struct
{
    int clockwise_deg_per_sec;
    int cclockwise_deg_per_sec;
    int down_deg_per_sec;
    int up_deg_per_sec;
    int current_x_degrees;
    int current_y_degrees;
} callibration = {0, 0, 0, 0};


/* Globals */
static struct usb_dev_handle *hnd = NULL;


static void usage(char **argv)
{
    printf(
"YeeHaa -- version %s\n"
"Copyright 2009 Matt Davis <mattdavis9@gmail.com>\n"
"yeehaa comes with ABSOLUTELY NO WARRANTY; for details see the COPYING text\n"
"located in the source directory of this application. This is free\n"
"software, and you are welcome to redistribute it under certain conditions;\n"
"see the COPYING text located in the source directory for this application\n"
"for details.\n\n", YEEHAA_VERSION);

    printf("Usage: %s [[-x <deg>] [-y <deg>]] OPTIONS\n"
           "       %s [-p <dir> <sec>] OPTIONS\n"
           "-x <deg>      : move deg degrees on the x-axis\n"
           "-y <deg>      : move deg degrees on the y-axis\n"
           "-p <dir> <sec>: move dir direction for sec seconds where\n"
           "                direction is <up | down | left | right>\n"
           "-s            : shoot a missile\n"
           "-i            : interactive mode (arrow keys and space to fire)\n"
           "-c            : recallibrate the device\n",
           argv[0], argv[0]);

    _Exit(0);
}


static char get_status(void)
{
    char stat;

    DO_CONTROL(_status);
    usb_bulk_read(hnd, 1, &stat, 1, 1000);

    return stat;
}


static void init_usb(void)
{
    struct usb_bus    *bb;
    struct usb_device *dd, *dev;

    usb_init();
    YEEHAA_ERR(usb_find_busses());
    YEEHAA_ERR(usb_find_devices());

    /* Find our launcher */
    dev = NULL;
    for (bb=usb_get_busses(); bb && !dev; bb=bb->next)
    {
        dev = NULL;
        for (dd=bb->devices; dd; dd=dd->next)
          if ((dd->descriptor.idVendor == VENDOR) &&
              (dd->descriptor.idProduct == PRODUCT))
          {
              dev = dd;
              break;
          }
    }

    YEEHAA_ERR(dev);
    YEEHAA_ERR(hnd = usb_open(dev));
    YEEHAA_ERR(!(usb_set_configuration(hnd, 1) < 0));
    YEEHAA_ERR(!(usb_claim_interface(hnd, 0) < 0));
}


static void deinit_usb(void)
{
    usb_release_interface(hnd, 0);
    usb_close(hnd);
}


static char get_position_limit_mask(char position)
{
    switch (position)
    {
        case POSITION_LEFT:  return LEFT_LIMIT;
        case POSITION_RIGHT: return RIGHT_LIMIT;
        case POSITION_UP:    return UP_LIMIT;
        case POSITION_DOWN:  return DOWN_LIMIT;
    }

    return POSITION_UNSET;
}

static void move_seconds(char position, float seconds, int do_callibrate)
{
    char   pos_limit;
    time_t st;

    /* Know when to stop */
    pos_limit = get_position_limit_mask(position);
    
    /* While we do not reach the limit and we click x seconds of movement */
    st = time(NULL);
    DO_CONTROL(position);

    if (!do_callibrate)
      while (!(get_status() & pos_limit) &&
             ((difftime(time(NULL), st) < seconds)))
      {
        ;
      }
    else /* We are callibrating so just go to extreme */
      while (!(get_status() & pos_limit))
      {
          ;
      }

    DO_CONTROL(_stop);
} 


static void move_turret(int degrees_x, int degrees_y)
{
    int  diff_x, diff_y;

    diff_x = diff_y = 0;

    /* Figure out if we move left (ccw) or right (clockwise) */
    if (degrees_x != IGNORE)
    {
        diff_x = callibration.current_x_degrees - degrees_x;

        /* Move left/right */
        if (diff_x > 0)
          move_seconds(POSITION_LEFT,
               abs(diff_x) / callibration.cclockwise_deg_per_sec,
               0);
        else if (diff_x < 0)
          move_seconds(POSITION_RIGHT, 
               abs(diff_x) / callibration.cclockwise_deg_per_sec,
               0);
    }

    /* Figure out if we move up or down */
    if (degrees_y != IGNORE)
    {
        diff_y = callibration.current_y_degrees - degrees_y;
        
        if (diff_y > 0)
          move_seconds(POSITION_DOWN,
                       abs(diff_y) / callibration.down_deg_per_sec,
                       0);
        else if (diff_y < 0)
          move_seconds(POSITION_UP,
                       abs(diff_y) / callibration.up_deg_per_sec,
                       0);
    }

    /* Save */
    callibration.current_x_degrees = degrees_x;
    callibration.current_y_degrees = degrees_y;
}


static void move_turret_seconds(char position, float seconds)
{
    /* Move */
    move_seconds(position, seconds, 0);

    /* Update */
    if (position == POSITION_LEFT)
      callibration.current_x_degrees -=
          callibration.cclockwise_deg_per_sec * seconds;
    else if (position == POSITION_RIGHT)
      callibration.current_x_degrees += 
          callibration.cclockwise_deg_per_sec * seconds;
    else if (position == POSITION_UP)
      callibration.current_y_degrees +=
          callibration.clockwise_deg_per_sec * seconds;
    else
      callibration.current_y_degrees -=
          callibration.clockwise_deg_per_sec * seconds;

    /* Save */
    callibration.current_x_degrees = callibration.current_x_degrees;
    callibration.current_y_degrees = callibration.current_y_degrees;

    /* Upper limit */
    if (callibration.current_x_degrees > 360)
      callibration.current_x_degrees = 360;
    if (callibration.current_y_degrees > 90)
      callibration.current_y_degrees = 90;
    
    /* Lower limit */
    if (callibration.current_x_degrees < 0)
      callibration.current_x_degrees = 0;
    if (callibration.current_y_degrees < 0)
      callibration.current_y_degrees = 0;
}


static void shoot(void)
{
    DO_CONTROL(_shoot);
    while (!(get_status() & HAS_SHOT))
      usleep(500000);
    DO_CONTROL(_stop);
}


static void calibrate(void)
{
    time_t st_cw, en_cw, st_ccw, en_ccw;
    time_t st_down, en_down, st_up, en_up;
    
    printf("-- Calibrating --\n");

    /* Move all the way to the left */
    printf("--> Resetting clockwise position\n");
    move_seconds(POSITION_LEFT, 0, 1);

    /* Calclate degrees per second of time (clockwise) */
    printf("--> Calculating clockwise degrees per second\n");
    st_cw = time(NULL);
    move_seconds(POSITION_RIGHT, 0, 1);
    en_cw = time(NULL);
    callibration.clockwise_deg_per_sec  = 360 / (int)difftime(en_cw, st_cw);
    printf("----> Clockwise Total Time        : %d seconds\n"
           "----> Clockwise Degrees Per Second: %d\n",
           (int)difftime(en_cw, st_cw),
           callibration.clockwise_deg_per_sec);

    /* Calclate degrees per second of time (counter-clockwise) */
    printf("--> Calculating counter clockwise degrees per second\n");
    st_ccw = time(NULL);
    move_seconds(POSITION_LEFT, 0, 1);
    en_ccw = time(NULL);
    callibration.cclockwise_deg_per_sec = 360 / (int)difftime(en_ccw, st_ccw);
    printf("----> Counter-Clockwise Total Time        : %d seconds\n"
           "----> Counter-Clockwise Degrees Per Second: %d\n",
           (int)difftime(en_ccw, st_ccw),
           callibration.clockwise_deg_per_sec);

    /* Setup current position (fully clockwise is '0') */
    callibration.current_x_degrees = 0;

    /* Calclate 0-90 for y axis going down */
    printf("--> Resetting upward position\n");
    move_seconds(POSITION_UP, 0, 1);
    printf("--> Calculating downward degrees per second\n");
    st_down = time(NULL);
    move_seconds(POSITION_DOWN, 0, 1);
    en_down = time(NULL);
    DO_CONTROL(_stop);
    callibration.down_deg_per_sec = 90 / (int)difftime(en_down, st_down);
    printf("----> Downward Total Time        : %d seconds\n"
           "----> Downward Degrees Per Second: %d\n",
           (int)difftime(en_down, st_down),
           callibration.down_deg_per_sec);
    
    /* Calclate 0-90 for y axis going up */
    printf("--> Calculating upward degrees per second\n");
    st_up = time(NULL);
    move_seconds(POSITION_UP, 0, 1);
    en_up = time(NULL);
    DO_CONTROL(_stop);
    callibration.up_deg_per_sec = 90 / (int)difftime(en_up, st_up);
    printf("----> Upward Total Time        : %d seconds\n"
           "----> Upward Degrees Per Second: %d\n",
           (int)difftime(en_up, st_up),
           callibration.up_deg_per_sec);

    /* Setup current y-position */
    callibration.current_y_degrees = 90;
}


#define TRY_UPDATE_CALLIBRATION(_n, _f, _v) \
    if (strstr(#_n, _f)) { callibration._n = _v; continue; }


static void load_config(const char *fname)
{
    int   value;
    char  *field, buf[256];
    FILE *fp;

    /* If the config does not exist, create it */
    if (!(fp = fopen(fname, "r")))
    {
        calibrate();
        return;
    }

    /* Load the data */
    while (fgets(buf, 256, fp))
    {
        if (strchr(buf, '#'))
          continue;
        field = strtok(buf, ":");
        value = atoi(strtok(NULL, ":"));
        TRY_UPDATE_CALLIBRATION(clockwise_deg_per_sec, field, value);
        TRY_UPDATE_CALLIBRATION(cclockwise_deg_per_sec, field, value);
        TRY_UPDATE_CALLIBRATION(down_deg_per_sec, field, value);
        TRY_UPDATE_CALLIBRATION(up_deg_per_sec, field, value);
        TRY_UPDATE_CALLIBRATION(current_x_degrees, field, value);
        TRY_UPDATE_CALLIBRATION(current_y_degrees, field, value);
    }
    fclose(fp);
}


#define ADD_ENTRY(_e) fprintf(fp, #_e": %d\n", callibration._e);


static void save_config(const char *fname)
{
    FILE *fp;
    
    /* If the config does not exist, create it */
    if (!(fp = fopen(fname, "w+")))
    {
        YEEHAA_ERR_MSG("Configuration file could not be created/opened");
        abort();
    }

    fprintf(fp, "# This is autogenerated, values can change but nothing \n"
                "# should be added and the field names should not be "
                "changed\n");
    ADD_ENTRY(clockwise_deg_per_sec);
    ADD_ENTRY(cclockwise_deg_per_sec);
    ADD_ENTRY(down_deg_per_sec);
    ADD_ENTRY(up_deg_per_sec);
    ADD_ENTRY(current_x_degrees);
    ADD_ENTRY(current_y_degrees);
    fclose(fp);
}


static char get_position(const char *str)
{
    if (str[0] == 'l' || str[0] == 'L')
      return POSITION_LEFT;
    else if (str[0] == 'r' || str[0] == 'R')
      return POSITION_RIGHT;
    else if (str[0] == 'u' || str[0] == 'U')
      return POSITION_UP;
    else if (str[0] == 'd' || str[0] == 'D')
      return POSITION_DOWN;
    else
      return IGNORE;
}


/* Listen for user input and act upon it */
static void interactive(void)
{
    int key;

    /* Initalize ncurses */
    initscr();
    cbreak();
    noecho();
    keypad(stdscr, TRUE);
    
    printw("Interactive mode:\n"
           "\tArrow keys move the turret\n"
           "\tEnter or Space Bar shoots a missile\n"
           "\tX or Q exits\n");

    while ((key = getch()))
    {
        switch (key)
        {
            case KEY_RIGHT: move_seconds(POSITION_RIGHT, 0.1, 0); break;
            case KEY_LEFT: move_seconds(POSITION_LEFT, 0.1, 0); break;
            case KEY_UP: move_seconds(POSITION_UP, 0.1, 0); break;
            case KEY_DOWN: move_seconds(POSITION_DOWN, 0.1, 0); break;

            case 32: /* Space bar */
            case KEY_ENTER:
                shoot();
                break;

            case 'x':
            case 'X':
            case 'q':
            case 'Q':
                endwin();
                return;
        }
        qiflush();
    }
}


int main(int argc, char **argv)
{
    int                i, x, y, sec;
    char               pos;
    unsigned           do_shoot, do_calibrate, do_interactive;

    if (argc == 1)
      usage(argv);

    x = y = pos = sec = IGNORE;
    do_shoot = do_calibrate = do_interactive = 0;

    for (i=1; i<argc; i++)
    {
        if (argv[i][0] != '-')
          usage(argv);
        else if (argv[i][1] == 's')
          do_shoot = 1;
        else if (argv[i][1] == 'c')
          do_calibrate = 1;
        else if (argv[i][1] == 'i')
          do_interactive = 1;
        else if (argv[i][1] == 'x' && strlen(argv[i]) == 2)
          x = atoi(argv[++i]);
        else if (argv[i][1] == 'x' && strlen(argv[i]) > 2)
          x = atoi(argv[i]+2);
        else if (argv[i][1] == 'y' && strlen(argv[i]) == 2)
          y = atoi(argv[++i]);
        else if (argv[i][1] == 'y' && strlen(argv[i]) > 2)
          y = atoi(argv[i]+2);
        else if (argv[i][1] == 'p')
        {
            if ((pos = get_position(argv[++i])) == IGNORE)
              usage(argv);
            if (!(sec = atoi(argv[++i])))
              usage(argv);
        }
        else
          usage(argv);
    }

    /* Made it this far, so we must be cool */
    init_usb();
    
    /* Initalize configuration */
    load_config(".yeehaa.config");

    /* Callibrate before trying any motions */
    if (do_calibrate)
      calibrate();

    /* Valid movement options */
    if ((sec == IGNORE) &&
        (((x != IGNORE) && (x>=0 && x<=360)) || 
         ((y != IGNORE) && (y>=0 && y<=90))))
      move_turret(x, y);
    else if ((sec != IGNORE) && (pos != IGNORE))
      move_turret_seconds(pos, sec);

    if (do_shoot)
      shoot();

    if (do_interactive)
      interactive();

    deinit_usb();
    save_config(".yeehaa.config");

    return 0;
}
